#ifndef _DBUTILS_DAQCONFIG_DCMHWCONFIG_H
#define _DBUTILS_DAQCONFIG_DCMHWCONFIG_H

#include <DatabaseUtils/DAQConfig/ASICRegisterValueSet.h>
#include <DatabaseUtils/DAQConfig/FEBPulserParameterSet.h>
#include <DatabaseUtils/DAQConfig/GlobalNameSpec.h>
#include <DatabaseUtils/DAQConfig/SubsystemNameSpec.h>
#include <boost/shared_ptr.hpp>
#include <boost/thread/mutex.hpp>
#include <iostream>
#include <string>
#include <vector>

namespace dbutils {
namespace daqconfig {

/**
 * Worker class to handle fetching and formatting of DCM
 * hardware configuration data.
 *
 * @author Kurt Biery
 * @version $Revision: 1.21 $ $Date: 2013/12/16 18:53:40 $
 */

class DCMHardwareConfig {

 public:

    // ********************
    // *** Constructors ***
    // ********************

    DCMHardwareConfig(const std::vector<int64_t>& subsystemConfigIdList,
                      const std::string& dcmName,
                      bool loadDataImmediately = true);

    DCMHardwareConfig(const SubsystemNameSpec& nameSpec,
                      const std::string& dcmName,
                      bool loadDataImmediately = true);

    DCMHardwareConfig(const GlobalNameSpec& nameSpec,
                      const std::string& dcmName,
                      bool loadDataImmediately = true);

    DCMHardwareConfig(const int64_t& globalConfigId,
                      const std::string& dcmName,
                      bool loadDataImmediately = true);

    virtual ~DCMHardwareConfig() {};

    // *******************
    // *** Initializer ***
    // *******************

    void loadData();

    // *****************
    // *** Accessors ***
    // *****************

    virtual bool hasValidData() const {
        boost::mutex::scoped_lock sl(_dataMutex);
        return _hasValidData;
    }
    int64_t getPrimarySubsystemConfigId() const {
        boost::mutex::scoped_lock sl(_dataMutex);
        return _primarySubsystemConfigId;
    }

    int getEnabledPixelCount();

 protected:

    bool _fetchNamedConfigList(const SubsystemNameSpec& nameSpec,
                               std::vector<int64_t>& idList);

    bool _fetchNamedConfigList(const GlobalNameSpec& nameSpec,
                               std::vector<int64_t>& idList);

    bool _fetchAllData(const std::vector<int64_t>& subsystemConfigIdList,
                       const std::string& dcmName);

    bool _fetchDCMSystemParams(const std::vector<int64_t>& subsystemConfigIdList);

    bool _fetchDCMFirmwareLocations(const std::vector<int64_t>& subsystemConfigIdList,
                                    const std::string& dcmName);
    bool _fetchFEBFirmwareLocations(const std::vector<int64_t>& subsystemConfigIdList,
                                    const std::string& dcmName);
    bool _fetchFEBEnableMasks(const std::vector<int64_t>& subsystemConfigIdList,
                              const std::string& dcmName);
    bool _fetchDCMDataDevValues(const std::vector<int64_t>& subsystemConfigIdList,
                                const std::string& dcmName);
    bool _fetchTimingSystemValues(const std::vector<int64_t>& subsystemConfigIdList,
                                  const std::string& dcmName);
    bool _fetchDCMFPGAParams(const std::vector<int64_t>& subsystemConfigIdList,
                             const std::string& dcmName);
    bool _fetchDCMApplicationParams(const std::vector<int64_t>& subsystemConfigIdList,
                                    const std::string& dcmName);

    bool _fetchPixelEnableMasks(const std::vector<int64_t>& subsystemConfigIdList,
                                const std::string& dcmName);
    bool _fetchASICRegisterValues(const std::vector<int64_t>& subsystemConfigIdList,
                                  const std::string& dcmName);
    bool _fetchDSODataRegulatorValues(const std::vector<int64_t>& subsystemConfigIdList,
                                      const std::string& dcmName);
    bool _fetchFEBPulserValues(const std::vector<int64_t>& subsystemConfigIdList,
                               const std::string& dcmName);

    bool _fetchPixelOffsets(const std::vector<int64_t>& subsystemConfigIdList,
                            const std::string& dcmName);
    bool _fetchPixelThresholds(const std::vector<int64_t>& subsystemConfigIdList,
                               const std::string& dcmName);

    std::string _dcmName;
    std::vector<int64_t> _subsystemConfigIdList;
    bool _hasValidData;
    int64_t _primarySubsystemConfigId;

    mutable boost::mutex _dataMutex;

    // system-wide parameters
    int32_t _microsliceTimeLength;
    std::string _dcmFPGAMode;

    // dcm-specific parameters
    std::string _dcmFirmwareLocation;
    int32_t _forceDCMFirmwareProgramming;
    std::string _febFirmwareLocation;
    int32_t _forceFEBFirmwareProgramming;
    int64_t _febEnableMask;
    int32_t _dcmReadTimeout;
    int32_t _useMemoryMappedBuffer;
    int16_t _timingSystemEnable;
    int16_t _timingSystemPort;
    int32_t _tduFirmwareVersion;
    int16_t _timingHistoryEnable;
    int32_t _fpgaLinkTimeout;
    int32_t _doDCMFPGAReset;
    int32_t _nanoSliceVersion;
    int16_t _numberOfHwConfigAttempts;

    // feb-specific parameters
    std::vector<int32_t> _pixelEnableList;
    std::vector<dbutils::daqconfig::ASICRegisterValueSet> _asicRegisterValueList;
    std::vector<int16_t> _dsoPointsPerPixelList;
    std::vector<dbutils::daqconfig::FEBPulserParameterSet> _pulserParameterList;

    // pixel-specific parameters
    std::vector<boost::shared_ptr<std::vector<int16_t> > > _pixelOffsetArray;
    std::vector<boost::shared_ptr<std::vector<int16_t> > > _pixelThresholdArray;

};

};  // end of namespace daqconfig
};  // end of namespace dbutil

#endif
