#ifndef __DBCOLUMN_HPP_
#define __DBCOLUMN_HPP_

#include <string>
#include <vector>

#include <boost/lexical_cast.hpp>

#include "NovaDatabase/NovaDatabase.h"

// using namespace std;

namespace nova {
  namespace database {

    enum ToleranceType {
      kUnknown,
      kAbsolute,
      kFractional,
      kNToleranceTypes
    };

    /**
     * Generalized Database Column Interface
     *
     * @author Jonathan Paley
     * @version $Id: Column.h,v 1.24 2013/03/12 19:53:02 jpaley Exp $
     */
    class Column //: public dBColumn_t
    {
    public:
      Column(const dBColumn_t& c);
      ~Column();
      
      std::string Name()      const { return fName;}
      std::string Type()      const { return fType;}
      std::string Value()     const { return fValue; }
      bool        CanBeNull() const { return fCanBeNull; }
      bool        IsNull()    const { return fIsNull; }
      bool        Modified()  const { return fModified; }
      int         ToleranceType() const { return fTolType; }
      float       Tolerance()   const { return fTolerance; }
      
      void        Clear();

      void        SetName(std::string n) { fName = n; }
      void        SetType(std::string t) { fType = t; }
      void        SetNull(bool isNull) { fIsNull = isNull; }
      void        SetToleranceType(int t) { fTolType = t; }
      void        SetTolerance(float t) { fTolerance = t; }

      template <class T>
	bool Get(T& val) const { 
	if (! fIsNull) {
	  try {
	    val = boost::lexical_cast<T>(fValue); 
	  }
	  catch (boost::bad_lexical_cast &) {
	    std::cerr << "Column::Get(): Bad_lexical_cast! Value = " 
		      << fValue << std::endl;
	    return false;
	  }
	  return true;
	}

	return false;
      }
    
      template <class T>
	bool Set(const T& val,bool ignoreAutoIncr=false) { 
	if (!ignoreAutoIncr && fType == "autoincr") {
	  std::cerr << "Cannot set a column of type \"autoincr\"!" 
		    << std::endl;
	  return false;
	}
	try {	  
	  std::string tstr = boost::lexical_cast<std::string>(val);
	  if (fType == "bool") {
	    fIsNull = false;
	    if (tstr == "TRUE" || tstr == "t" || tstr == "true" || 
		tstr == "y" || tstr == "yes" || tstr == "1" || tstr == "on")
	      fValue = "1";
	    else
	      fValue = "0";
	    return true;
	  }
	  else {
	    if (tstr == "") return false;
	    fIsNull = false;
	    if (tstr == "NULL") 
	      fIsNull = true;
	    fValue = tstr;
	    return true;
	  }
	}
	catch (boost::bad_lexical_cast &) {
	  std::cerr << "Column::Set(): Bad lexical cast! Value = " 
	       << val << std::endl;
	  return false;
	}
	return false;
      }

      template <class T>
	bool Update(const T& val) { 
	if (Set(val)) {
	  fModified = true; return true; }
	else
	  return false;
      }

      friend std::ostream& operator<< (std::ostream& stream, const Column& col);
	
      bool        operator >= (const Column& c) const;
      bool        operator <= (const Column& c) const;
      bool        operator > (const Column& c) const;
      bool        operator < (const Column& c) const;
      bool        operator == (const Column& c) const;

    private:
      bool        fCanBeNull;
      bool        fIsNull;
      bool        fModified;
      int         fTolType;
      float       fTolerance;
      std::string fType;
      std::string fName;
      std::string fValue;

    }; // class end

    //************************************************************
    
    inline std::ostream& operator<< (std::ostream& stream, const Column& col) { 
      if (col.fIsNull) {
	stream << "NULL";
      }
      else {
	if (col.fType == "bool") {
	  if (col.fValue == "1")
	    stream << "true";
	  else
	    stream << "false";
	}
	else {
	  bool needsQuotes = (col.fType == "string" || col.fType == "text" || 
			      col.fType == "timestamp" || col.fType == "date" );
	  if (needsQuotes)	stream << "\'";
	  stream << col.fValue;
	  if (needsQuotes)	stream << "\'";
	}
      }
      return stream;
    }

  } // namespace novadaqdatabase close
} // namespace nova close

#endif
