/* 
 * File:   QualityCheck.cxx
 * Author: mmathis
 * 
 * Created on June 23, 2010, 12:24 PM
 */
#include <stdio.h>
#include <vector>
#include "DAQQualityCheck/QualityCheck.h"
#include "DAQDataFormats/RawEvent.h"
#include "DAQDataFormats/RawDataBlock.h"
#include "DAQDataFormats/RawTriggerHeader.h"
#include "DAQDataFormats/RawTriggerMask.h"
#include <inttypes.h>

using namespace daqqc;


//Constructors
QualityCheck::QualityCheck() {
  fCheckRun = false;
  fEventCounter = 0;
}


//Destructor
QualityCheck::~QualityCheck() {
}


#ifndef RAWRUN_IS_NOT_IMPLEMENTED
/**
 * Runs all the desired consistency checks on the RawRun. Returns a single
 * boolean indicating if all tests were successful, or if there was at least
 * one problem in at least one of the checks. Individual problems are printed
 * to the standard output and/or a specified buffer.
 */
bool QualityCheck::checkRun(daqdataformats::RawRun& run) {
  bool result = true, temp = true;;
  fRun = run;
  fCheckRun = true;
  std::cout << "Running some checks on the run" << std::endl;

  fPartition = fRun.getPartition();
  fRunNum = fRun.getRunNumber();
  fSubrunNum = fRun.getSubrun();
  fSim = fRun.getTriggerMC();
  fEventCounter = fRun.getUniversalEventStart();

  // Check the CRC for the entire run
  if (fRun.calculateCheckSum(0) != 0) {
    result = false;
    std::cout << "QC ERROR: Run CRC checksum is wrong" << std::endl;
  }

  temp = checkRunHeaderTail(fRun.getHeader(), fRun.getTail());
  if (!result && temp)
    result = false;
  else
    result = temp;

  temp = checkConfigBlock(fRun.getConfigurationBlock());
  if (!result && temp)
    result = false;
  else
    result = temp;

  // Run checks on each event
  for (uint32_t ievt = 0; ievt < fRun.getTotalEvents(); ievt++) {
    fRun.setFloatingEvent(ievt);
    daqdataformats::RawEvent* event = fRun.getFloatingEvent();

    if (ievt == 0) {
      if (event->getTrigger()->getTriggerTimingMarker()->getTriggerStart_baseClock() != fRun.getHeader()->getRunStartTime()) {
	result = false;
	std::cout << "QC ERROR: Run start time not consistent with first event's trigger time" << std::endl;
      }
    } // if ievt = 0
    if (ievt == fRun.getTotalEvents() - 1) {
      if (event->getTrigger()->getTriggerTimingMarker()->getTriggerStart_baseClock() != fRun.getHeader()->getRunEndTime()) {
	result = false;
	std::cout << "QC ERROR: Run end time not consistent with last event's trigger time" << std::endl;
      }
    } // if ievt

    temp = checkEvent(event);    
    if (!result && temp)
      result = false;
    else
      result = temp;

    fEventCounter++; 
  }

  return result;
} // checkRun
#endif

/**
 * Runs all the desired consistency checks on the RawEvent. Returns a single
 * boolean indicating if all tests were successful, or if there was at least
 * one problem in at least one of the checks. The return value does not 
 * specify what went wrong, only that something did go wrong. Individual 
 * problems are printed to the standard output and/or a specified buffer.
 * @param event The event to check
 */
bool QualityCheck::checkEvent(daqdataformats::RawEvent& event) {
  bool result = true, temp = true;
  fEvent = event;

  std::cout << "\nChecking the event" << std::endl;

  // First, do the CRC checksum for the entire event
  if (fEvent.isCRCCalculationUsed()) {
    if (fEvent.calculateCheckSum(0) != 0) {
      result = false;
      std::cout << "QC ERROR: Event CRC checksum wrong" << std::endl;
    }
  }

  fEventNum = event.getEventNumber();
  fTrigger = event.getTrigger()->getTriggerHeader()->getTriggerNumber();

  if (fCheckRun) {
    if (fEventNum != fEventCounter) {
      result = false;
      std::cout << "QC ERROR: Event number did not increment correctly" << std::endl;
    }
  } else {
    fPartition = event.getPartition();
    fSim = event.getMonteCarloFlag();
  }

  temp = checkTriggerBlock(event.getTrigger());
  if (!result && temp)
    result = false;
  else
    result = temp;

  temp = checkEventHeaderTail(event.getHeader(), event.getTail());
  if (!result && temp)
    result = false;
  else
    result = temp;

  // Loop over the datablocks
  uint32_t ndata = event.getDataBlockNumber();
  for (uint32_t idata = 0; idata < ndata; idata++) {
    event.setFloatingDataBlock(idata);
    daqdataformats::RawDataBlock* data = event.getFloatingDataBlock();

    temp = checkDataBlock(data, idata);
    if (!result && temp)
      result = false;
    else
      result = temp;

    // Loop over micro-blocks
    uint16_t nmicrob = data->getNumMicroBlocks();
    for (uint16_t imicrob = 0; imicrob < nmicrob; imicrob++) {
      data->setFloatingMicroBlock(imicrob);
      daqdataformats::RawMicroBlock*       microblock = (daqdataformats::RawMicroBlock*)data->getFloatingMicroBlock();
      daqdataformats::RawMicroBlockHeader* microhead  = (daqdataformats::RawMicroBlockHeader*)microblock->getHeader();
      // Check partition in Microblock header
      if (microhead->getPartition() != fPartition) {
	result = false;
	std::cout << "QC ERROR: Datablock " << idata << ", Microblock " << imicrob << " partition wrong: " << microhead->getPartition() << std::endl;
      }

      // There is only 1 micro slice per micro block
      daqdataformats::RawMicroSlice* micro = microblock->getMicroSlice();
      temp = checkMicroSlice(micro, idata, imicrob);
      if (!result && temp)
	result = false;
      else
	result = temp;

      // Loop over nanoslices in this microslice   
      uint32_t nnano = micro->getNumNanoSlices();
      for (uint32_t inano = 0; inano < nnano; inano++) {
	micro->setFloatingNanoSlice(inano);
	daqdataformats::RawNanoSlice* nano = micro->getFloatingNanoSlice();
	temp = checkNanoSlice(nano);
	if (!result && temp)
	  result = false;
	else
	  result = temp;
	
      } // loop over nanoslices
    } // loop over microblocks
  } // loop over datablocks

  return result;
} // runEventChecks


/**
 * Helper function to run all the desired consistency checks on the 
 * RawEvent. Returns a single
 * boolean indicating if all tests were successful, or if there was at least
 * one problem in at least one of the checks. 
 * @param event The event to check
 */
bool QualityCheck::checkEvent(daqdataformats::RawEvent* event) {
  bool result = true, temp = true;

  std::cout << "Checking the event" << std::endl;

  // First, do the CRC checksum for the entire event
  if (event->isCRCCalculationUsed()) {
    if (event->calculateCheckSum(0) != 0) {
      result = false;
      std::cout << "QC ERROR: Event CRC checksum wrong" << std::endl;
    }
  }

  fEventNum = event->getEventNumber();
  fTrigger = event->getTrigger()->getTriggerHeader()->getTriggerNumber();

  if (fCheckRun) {
    if (fEventNum != fEventCounter) {
      result = false;
      std::cout << "QC ERROR: Event number did not increment correctly" << std::endl;
    }
  } else {
    fPartition = event->getPartition();
    fSim = event->getMonteCarloFlag();
  }

  temp = checkTriggerBlock(event->getTrigger());
  if (!result && temp)
    result = false;
  else
    result = temp;
  temp = checkEventHeaderTail(event->getHeader(), event->getTail());
  if (!result && temp)
    result = false;
  else
    result = temp;

  // Loop over the data-blocks
  uint32_t ndata = event->getDataBlockNumber();
  for (uint32_t idata = 0; idata < ndata; idata++) {
    event->setFloatingDataBlock(idata);
    daqdataformats::RawDataBlock* data = event->getFloatingDataBlock();

    temp = checkDataBlock(data, idata);
    if (!result && temp)
      result = false;
    else
      result = temp;

    // Loop over micro-blocks
    uint16_t nmicrob = data->getNumMicroBlocks();
    for (uint16_t imicrob = 0; imicrob < nmicrob; imicrob++) {
      data->setFloatingMicroBlock(imicrob);
      daqdataformats::RawMicroBlock*       microblock = (daqdataformats::RawMicroBlock*)data->getFloatingMicroBlock();
      daqdataformats::RawMicroBlockHeader* microhead  = (daqdataformats::RawMicroBlockHeader*)microblock->getHeader();
      // Check partition in Microblock header
      if (microhead->getPartition() != fPartition) {
	result = false;
	std::cout << "QC ERROR: Datablock " << idata << ", Microblock " << imicrob << " partition wrong: " << microhead->getPartition() << std::endl;
      }

      // There is only 1 micro slice per micro block
      daqdataformats::RawMicroSlice* micro = microblock->getMicroSlice();
      temp = checkMicroSlice(micro, idata, imicrob);
      if (!result && temp)
	result = false;
      else
	result = temp;

      // Loop over nanoslices in this microslice   
      uint32_t nnano = micro->getNumNanoSlices();
      for (uint32_t inano = 0; inano < nnano; inano++) {
	micro->setFloatingNanoSlice(inano);
	daqdataformats::RawNanoSlice* nano = micro->getFloatingNanoSlice();
	temp = checkNanoSlice(nano);
	if (!result && temp)
	  result = false;
	else
	  result = temp;

      } // loop over nanoslices
    } // loop over microblocks
  } // loop over datablocks

  return result;
} // runEventChecks


/**
 * Helper to check that the run header and tail are consistent
 * with each other
 * @param header The run header
 * @param tail The run tail. Note that it is of type RawRunHeader, as they 
 * have the same structure
 */
bool QualityCheck::checkRunHeaderTail(daqdataformats::RawRunHeader* header, daqdataformats::RawRunHeader* tail) {
  bool result = true;
  std::cout << "Checking the run header and tail" << std::endl;

  // Check markers in header
  if (header->getMarkerHi() != 0x41764F4E) {
    result = false;
    std::cout << "QC ERROR: Run header high marker NOvA is wrong" << std::endl;
  }
  if (header->getMarkerLo() != 0x39323945) {
    result = false;
    std::cout << "QC ERROR: Run header low marker E929 is wrong" << std::endl;
  }

  // Check markers in tail
  if (tail->getMarkerHi() != 0x4C494154) {
    result = false;
    std::cout << "QC ERROR: Run tail upper high marker TAIL is wrong" << std::endl;
  }
  if (tail->getMarkerLo() != 0x4C494154) {
    result = false;
    std::cout << "QC ERROR: Run tail upper low marker TAIL is wrong" << std::endl;
  }
  if (tail->getReserved1() != 0x39323945) {
    result = false;
    std::cout << "QC ERROR: Run tail lower high marker E929 is wrong" << std::endl;
  }
  if (tail->getReserved2() != 0x41764F4E) {
    result = false;
    std::cout << "QC ERROR: Run tail lower low marker NOvA is wrong" << std::endl; 
  }
 
  // Check that various quantities are consistent between header and tail
  if (header->getRunNumber() != tail->getRunNumber()) {
    result = false;
    std::cout << "QC ERROR: Run number not consistent between run header and tail" << std::endl;
  }
  if (header->getRunType() != tail->getRunType()) {
    result = false;
    std::cout << "QC ERROR: Run type not consistent between run header and tail" << std::endl;
  }
  if (header->getSubrun() != tail->getSubrun()) {
    result = false;
    std::cout << "QC ERROR: Subrun number not consistent between run header and tail" << std::endl;
  }
  if (header->getPartition() != tail->getPartition()) {
    result = false;
    std::cout << "QC ERROR: Partition not consistent between run header and tail" << std::endl;
  }
  if (header->getRunControlVersion() != tail->getRunControlVersion()) {
    result = false;
    std::cout << "QC ERROR: Run control version not consistent between run header and tail" << std::endl;
  }
  if (header->getRunControlID() != tail->getRunControlID()) {
    result = false;
    std::cout << "QC ERROR: Run control ID not consistent between run header and tail" << std::endl;
  }
  if (header->getDataLoggerVersion() != tail->getDataLoggerVersion()) {
    result = false;
    std::cout << "QC ERROR: Datalogger version not consistent between run header and tail" << std::endl;
  }
  if (header->getDataLoggerID() != tail->getDataLoggerID()) {
    result = false;
    std::cout << "QC ERROR: Datalogger ID not consistent between run header and tail" << std::endl;
  }
  if (header->getRunStartTime() != tail->getRunStartTime()) {
    result = false;
    std::cout << "QC ERROR: Run start time not consistent between run header and tail" << std::endl;
  }
  if (header->getRunEndTime() != tail->getRunEndTime()) {
    result = false;
    std::cout << "QC ERROR: Run end time not consistent between run header and tail" << std::endl;
  }
  if (header->getSubRunStartTime() != tail->getSubRunStartTime()) {
    result = false;
    std::cout << "QC ERROR: Subrun start time not consistent between run header and tail" << std::endl;
  }
  if (header->getSubRunEndTime() != tail->getSubRunEndTime()) {
    result = false;
    std::cout << "QC ERROR: Subrun end time not consistent between run header and tail" << std::endl;
  }
  if (header->getUniversalEventStart() != tail->getUniversalEventStart()) {
    result = false;
    std::cout << "QC ERROR: Universal event start not consistent between run header and tail" << std::endl;
  }
  if (header->getUniversalEventEnd() != tail->getUniversalEventEnd()) {
    result = false;
    std::cout << "QC ERROR: Universal event end not consistent between run header and tail" << std::endl;
  }
  if (header->getRunSize() != tail->getRunSize()) {
    result = false;
    std::cout << "QC ERROR: Run size not consistent between run header and tail" << std::endl;
  }
  if (header->getTriggerCtrlID() != tail->getTriggerCtrlID()) {
    result = false;
    std::cout << "QC ERROR: Trigger control ID not consistent between run header and tail" << std::endl;
  }
  if (header->getTriggerVersion() != tail->getTriggerVersion()) {
    result = false;
    std::cout << "QC ERROR: Trigger version not consistent between run header and tail" << std::endl;
  }
  if (header->getTriggerMC() != tail->getTriggerMC()) {
    result = false;
    std::cout << "QC ERROR: SIM flag not consistent between run header and tail" << std::endl;
  }
  if (header->getVTMod() != tail->getVTMod()) {
    result = false;
    std::cout << "QC ERROR: VT Mod not consistent between run header and tail" << std::endl;
  }
  if (header->getTriggerPrescaleListIDX() != tail->getTriggerPrescaleListIDX()) {
    result = false;
    std::cout << "QC ERROR: Trigger prescale list IDX not consistent between run header and tail" << std::endl;
  }
  if (header->getTriggerListIDX() != tail->getTriggerListIDX()) {
    result = false;
    std::cout << "QC ERROR: Trigger list IDX not consistent between run header and tail" << std::endl;
  }

  if ((header->getValidTriggerTypesLow() != tail->getValidTriggerTypesLow()) &&
      (header->getValidTriggerTypesHigh() != tail->getValidTriggerTypesHigh()) &&
      (header->getValidTriggerTypesHigh2() != tail->getValidTriggerTypesHigh2())) {
    result = false;
    std::cout << "QC ERROR: Valid trigger types not consistent between run header and tail" << std::endl;
  }
  
  // Check that the num events = event end - event start + 1
  if (header->getUniversalEventEnd() - header->getUniversalEventStart() + 1 != header->getTotalEvents()) {
    result = false;
    std::cout << "QC ERROR: Run header number of events not consistent with start and end event numbers" << std::endl;
  }

  // Check the CRC checksums in the header and tail
  if (header->calculateCheckSum(0) != 0) {
    result = false;
    std::cout << "QC ERROR: Run header checksum failed" << std::endl;
  }

  if (tail->calculateCheckSum(0) != 0) {
    result = false;
    std::cout << "QC ERROR: Run tail checksum failed" << std::endl;
  }

  // Check to make sure some values are non-zero
  if (header->getRunControlVersion() == 0) {
    result = false;
    std::cout << "QC ERROR: Run header Run control version is zero (0)" << std::endl;
  }
  if (header->getRunControlID() == 0) {
    result = false;
    std::cout << "QC ERROR: Run header Run control ID is zero (0)" << std::endl;
  }
  if (header->getDataLoggerVersion() == 0) {
    result = false;
    std::cout << "QC ERROR: Run headerDatalogger version is zero (0)" << std::endl;
  }
  if (header->getDataLoggerID() == 0) {
    result = false;
    std::cout << "QC ERROR: Run header Datalogger ID is zero (0)" << std::endl;
  }
  if (header->getRunSize() == 0) {
    result = false;
    std::cout << "QC ERROR: Run header run size is zero (0)" << std::endl;
  }
  if (header->getTriggerCtrlID() == 0) {
    result = false;
    std::cout << "QC ERROR: Run header trigger control ID is zero (0)" << std::endl;
  }
  if (header->getTriggerVersion() == 0) {
    result = false;
    std::cout << "QC ERROR: Run header trigger version is zero (0)" << std::endl;
  }
  /*if (header->getVTMod() == 0) {
    result = false;
    std::cout << "QC ERROR: Run header VT Mod is zero (0)" << std::endl;
    }*/
  if (header->getTriggerPrescaleListIDX() == 0) {
    result = false;
    std::cout << "QC ERROR: Run header trigger prescale list IDX is zero (0)" << std::endl;
  }
  if (header->getTriggerListIDX() == 0) {
    result = false;
    std::cout << "QC ERROR: Run header trigger list IDX is zero (0)" << std::endl;
  }

  return result;
} // checkRunHeaderTail


/**
 * Helper to check that the config block is consistent with the run header
 * @param config The Configuration Block
 */
bool QualityCheck::checkConfigBlock(daqdataformats::RawConfigurationBlock* config) {
  bool result = true;
  std::cout << "Checking the configuration block" << std::endl;
  //config->print();

  // First, ckeck the CRC checksum
  if (config->calculateCheckSum(0) != 0) {
    result = false;
    std::cout << "QC ERROR: Config block checksum failed" << std::endl;
  }

  // Now check the markers
  if (config->getMarkerHi1() != 0x464E4F43) {
    result = false;
    std::cout << "QC ERROR: Config block first marker CONF is wrong" << std::endl;
  }
  if (config->getMarkerHi2() != 0x54525453) {
    result = false;
    std::cout << "QC ERROR: Config block second marker STRT is wrong" << std::endl;
  }
  if (config->getMarkerLo1() != 0x464E4F43) {
    result = false;
    std::cout << "QC ERROR: Config block third marker CONF is wrong" << std::endl;
  }
  if (config->getMarkerLo2() != 0x44444E45) {
    result = false;
    std::cout << "QC ERROR: Config block fourth marker ENDD is wrong: " << config->getMarkerLo2() << ", " << config->getMarkerLo1() << std::endl;
  }

 #ifndef RAWRUN_IS_NOT_IMPLEMENTED    
   // Check run number, type, etc are consistent with run header
  if (config->getRunType() != fRun.getRunType()) {
    result = false;
    std::cout << "QC ERROR: Config block run type is not consistent with run header" << std::endl;
  }
#endif  
  if (config->getRunNumber() != fRunNum) {
    result = false;
    std::cout << "QC ERROR: Config block run number is not consistent with run header" << std::endl;
  }
  if (config->getSubrun() != fSubrunNum) {
    result = false;
    std::cout << "QC ERROR: Config block subrun number is not consistent with run header" << std::endl;
  }
  if (config->getPartition() != fPartition) {
    result = false;
    std::cout << "QC ERROR: Config block partition number is not consistent with run header" << std::endl;
  }
  /* if (config->getConfigIDX() != fRun.getConfigIDX()) {
    result = false;
    std::cout << "QC ERROR: Config block Config IDX is not consistent with run header" << std::endl;
  }
  */
  // Check to make sure some values are non-zero
  if (config->getConfigurationVersion() == 0) {
    result = false;
    std::cout << "QC ERROR: Config block header version number is zero (0)" << std::endl;
  }
  if (config->getResourceManagerID() == 0) {
    result = false;
    std::cout << "QC ERROR: Config block resource manager ID is zero (0)" << std::endl;
  }
  if (config->getSystemVersionIDX() == 0) {
    result = false;
    std::cout << "QC ERROR: Config block system version IDX is zero (0)" << std::endl;
  }
  if (config->getSystemNumber() == 0) {
    result = false;
    std::cout << "QC ERROR: Config block number of systems is zero (0)" << std::endl;
  }
  /*  if (config->getConfigIDX() == 0) {
    result = false;
    std::cout << "QC ERROR: Config block config IDX is zero (0)" << std::endl;
    }*/
  // Check ID and version numbers of all systems to make sure they are non-zero
  for (uint32_t i = 1; i < config->getSystemNumber(); i++) {
    if (config->getSystemID(i) == 0) {
      result = false;
      std::cout << "QC ERROR: Config block DAQ system ID " << i << " is zero (0)" << std::endl;
    }
  }

  return result;
} // checkConfigBlock


/**
 * Helper to check that the event header and tail are consistent 
 * with each other
 * @param header The event header
 * @param tail The event tail
 */
bool QualityCheck::checkEventHeaderTail(daqdataformats::RawEventHeader* header, daqdataformats::RawEventTail* tail) {
  bool result = true;

  std::cout << "Checking event header/tail:" << std::endl;

  // Check markers in the header
  if (header->getMarkerLeft() != 0xE929) {
    result = false;
    std::cout << "QC ERROR: Event header 0xE929 left marker wrong" << std::endl;
  }

  if (header->getMarkerRight() != 0xE929) {
    result = false;
    std::cout << "QC ERROR: Event header 0xE929 right marker wrong" << std::endl;
  }

  // Check markers in the tail
  if (tail->getMarker() != 0xEEEE) {
    result = false;
    std::cout << "QC ERROR: Event tail 0xEEEE marker wrong" << std::endl;
  }

  if (tail->getMarkerLeft() != 0x929E) {
    result = false;
    std::cout << "QC ERROR: Event tail 0x929E left marker wrong" << std::endl;
  }

  if (tail->getMarkerRight() != 0x929E) {
    result = false;
    std::cout << "QC ERROR: Event tail 0x929E right marker wrong" << std::endl;
  }

  // Check version number is non-zero in header
  if (header->getVersion() == 0) {
    result = false;
    std::cout << "QC ERROR: Event header version number is zero (0)" << std::endl;
  }

  
  // If part of a run, check some values are consistent with run, or check
  // that they're non-zero
  if (fCheckRun) {
#ifndef RAWRUN_IS_NOT_IMPLEMENTED          
    if (header->getDataLoggerID() != fRun.getDataLoggerID()) {
      result = false;
      std::cout << "QC ERROR: Event header datalogger ID is not consistent with run header" << std::endl;
    }
    if (header->getRunType() != fRun.getRunType()) {
      result = false;
      std::cout << "QC ERROR: Event header run type is not consistent with run header" << std::endl;
    }
    if (header->getConfigIDX() != fRun.getConfigIDX()) {
      result = false;
      std::cout << "QC ERROR: Event header ConfigIDX is not consistent with run header" << std::endl;
    }
#endif    
    if (header->getRunNumber() != fRunNum) {
      result = false;
      std::cout << "QC ERROR: Event header run number is not consistent with run header" << std::endl;
    }
    if (header->getSubrun() != fSubrunNum) {
      result = false;
      std::cout << "QC ERROR: Event header subrun number is not consistent with run header" << std::endl;
    }

  } else { // check that some values are non-zero
    if (header->getDataLoggerID() == 0) {
      result = false;
      std::cout << "QC ERROR: Event header datalogger ID is zero (0)" << std::endl;
    }
    if (header->getConfigIDX() == 0) {
      result = false;
      std::cout << "QC ERROR: Event header ConfigIDX is zero (0)" << std::endl;
    }
  } // if fCheckRun

  // Check event number and event size consistent
  uint64_t tailEventNum;
  uint32_t low, high;

  low  = tail->getEventNumberLo();
  high = tail->getEventNumberHi();

  tailEventNum = (uint64_t)high << 32 | low;

  if (tailEventNum != header->getEventNumber()) {
    result = false;
    std::cout << "QC ERROR: Event tail event number wrong: " << tailEventNum << std::endl;
  }

  if (tail->getEventSize() != header->getEventSize()) {
    result = false;
    std::cout << "QC ERROR: Event tail event size wrong:" << tail->getEventSize() << std::endl;
  }

  return result;
} // checkEventHeaderTail


/**
 * Helper to perform checks on the Trigger block
 * @param trigger The trigger block to check
 */
bool QualityCheck::checkTriggerBlock(daqdataformats::RawTrigger* trigger) {
  bool result = true;
  std::cout << "Check trigger" << std::endl;
  
  // Check marker in header
  if (trigger->getTriggerHeader()->getDelimiter() != 0xAAAA) {
    result = false;
    std::cout << "QC ERROR: Trigger header marker wrong" << std::endl;
  }

  // Check SIM flag in trigger mask
  if (trigger->getTriggerMask()->getMonteCarloFlag() != fSim) {
    result = false;
    std::cout << "QC ERROR: Trigger Mask SIM flag wrong: " << trigger->getTriggerMask()->getMonteCarloFlag() <<  std::endl;
  }

  // Check VTMod and Trig Types, compare to run header, if applicable
  if (fCheckRun) {
#ifndef RAWRUN_IS_NOT_IMPLEMENTED          
    if (trigger->getTriggerMask()->getTrigMod() != fRun.getVTMod()) {
      result = false;
      std::cout << "QC ERROR: Trigger Mask Trigger Mod not consistent with VTMod in run header" << std::endl;
    }
    if ((trigger->getTriggerMask()->getTrigMask_Low() != fRun.getValidTriggerTypesLow()) 
	|| (trigger->getTriggerMask()->getTrigMask_High() != fRun.getValidTriggerTypesHigh()) 
	|| (trigger->getTriggerMask()->getTrigMask_High2() != fRun.getValidTriggerTypesHigh2())) {
      result = false;
      std::cout << "QC ERROR: Trigger mask trigger types not consistent with run header" << std::endl;
    }
#endif    
  } else {


  } // if fCheckRun

  return result;
} // checkTriggerBlock


/**
 * Helper to perform checks on the DataBlock.
 * @param data The data block to check
 * @param idata The data block number, used only for printouts
 */
bool QualityCheck::checkDataBlock(daqdataformats::RawDataBlock* data, uint32_t idata) {
  bool result = true;
  std::cout << "Check data block" << std::endl;
  
  // Check version, ID, etc are non-zero
  if (data->getVersion() == 0) {
    result = false;
    std::cout << "QC ERROR: Datablock " << idata << " version number is zero (0)" << std::endl;
  }
  /*if (data->getBuffId() == 0) {
    result = false;
    std::cout << "QC ERROR: Datablock " << idata << " buffer node ID is zero (0)" << std::endl;
    }*/

  // Check trigger number
  if (data->getTrigNum() != fTrigger) {
    result = false;
    std::cout << "QC ERROR: Datablock " << idata << " trigger number wrong:" 
	      << fTrigger << " vs " << data->getTrigNum() << std::endl;
  }
  
  // Check SIM flag
  if (data->getMonteCarloFlag() != fSim) {
    result = false;
    std::cout << "QC ERROR: Datablock " << idata << " SIM flag wrong" << std::endl;
  }
  
  // Check 0xDABC marker
  if (data->getMarker() != 0xDABC) {
    result = false;
    std::cout << "QC ERROR: Datablock " << idata << " 0xDABC marker wrong" << std::endl;
  }

  // Check CRC checksum
  if (data->isCRCCalculationUsed()) {
    if (data->calculateCheckSum(0) != 0) {
      result = false;
      std::cout << "QC ERROR: Datablock " << idata << " CRC checksum wrong" << std::endl;
    }
  }
  
  return result;
} // checkDataBlock


/**
 * Helper to perform checks on the MicroSlice.
 * @param micro The MicroSlice to check
 * @param idata The data block the MicroSlice belongs to, used only
 * in printouts
 * @param imicrob The MicroBlock the MicroSlice belongs to, used only 
 * in printouts
 */
bool QualityCheck::checkMicroSlice(daqdataformats::RawMicroSlice* micro, uint32_t idata, uint32_t imicrob) {
  bool result = true;
  std::cout << "Check MicroSlice" << std::endl;
  
  // Check SIM flag
  if (micro->getMonteCarloFlag() != fSim) {
    result = false;
    std::cout << "QC ERROR: Datablock " << idata << ", Microblock " << imicrob << " SIM flag wrong" << std::endl;
  }
  
  // Check the DP flag
  bool dp = micro->getDataPresent();
  uint32_t nnano = micro->getNumNanoSlices();
  if ((dp && nnano == 0) ||
      (!dp && nnano != 0)) {
    result = false;
    std::cout << "QC ERROR: Datablock " << idata << ", Microblock " << imicrob << " DP flag wrong via num NanoSlices" << std::endl;
  }

  if ((dp && micro->getByteCount() <= 8) || (!dp && micro->getByteCount() > 8)) {
    result = false;
    std::cout << "QC ERROR: Datablock " << idata << ", Microblock " << imicrob << " DP flag wrong via header byte count" << std::endl;
  }

  return result;
} // checkMicroSlice


/**
 * Helper to perform checks on the NanoSlice.
 * @param nano The NanoSlice to check.
 */
bool QualityCheck::checkNanoSlice(daqdataformats::RawNanoSlice* nano) {
  bool result = true;
  std::cout << "Check NanoSlice" << std::endl;

  // Check the DP flag
  bool dp = nano->getHeader()->getDataPresent();
  if ((dp && nano->getPulseHeight() == 0) || (!dp && nano->getPulseHeight() != 0)) {
    result = false;
    std::cout << "QC ERROR: Nanoslice DP flag wrong: " << dp << ", " << nano->getPulseHeight() << std::endl;
  }

  // Check the FEB Status against SIM flag
  uint32_t stat = nano->getHeader()->getFEBStatus();

  if ((stat <= 0x3F && fSim) || (stat >= 0x40 && !fSim)) {
    result = false;
    std::cout << "QC ERROR: Nanoslice 'SIM flag' wrong:" << stat << " vs " << fSim << std::endl;
  }

  return result;
} // checkNanoSlice



