#include <NovaDAQUtilities/CacheManager.h>

namespace novadaq {

/**
 * Returns the single instance of the CacheManager.
 *
 * @returns CacheManager instance.
 */
CacheManager& CacheManager::getInstance() {
    static CacheManager singletonInstance;
    return singletonInstance;
}

/**
 * Registers the specified cache with the CacheManager.  Registered
 * caches will be flushed when the CacheManager receives a flush
 * request.
 *
 * @param cache a pointer to the cache that should be registered.
 */
void CacheManager::registerCache(Cache* cache) {
    boost::mutex::scoped_lock lock(_registryMutex);
    _registry.push_back(cache);
}

/**
 * Removes the specified cache from the CacheManager.
 *
 * @param cache a pointer to the cache that should be removed.
 */
void CacheManager::removeCache(Cache* cache) {
    boost::mutex::scoped_lock lock(_registryMutex);
    _registry.erase(remove(_registry.begin(), _registry.end(), cache),
                    _registry.end());
}

/**
 * Flushes the caches that have been registered.
 */
void CacheManager::flushCaches() {
    boost::mutex::scoped_lock lock(_registryMutex);
    for (uint32_t idx = 0; idx < _registry.size(); ++idx) {
        _registry[idx]->flush();
    }
}

};  // end of namespace novadaq
