#include <NovaDAQUtilities/HexUtils.h>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <sstream>
#include <iomanip>

namespace novadaq {
namespace HexUtils {

/**
 * Formats the specified 32-bit integer as a hex string with
 * the specified number of hex digits.
 */
std::string format(const int32_t& value, const int& ndigits)
{
    std::ostringstream formatter;
    formatter << "0x" << std::hex << std::setfill('0')
              << std::setw(ndigits) << std::uppercase << value;
    return formatter.str();
}

/**
 * Formats the specified 32-bit integer as a hex string with
 * the specified number of hex digits.
 */
std::string format(const int64_t& value, const int& ndigits)
{
    std::ostringstream formatter;
    formatter << "0x" << std::hex << std::setfill('0')
              << std::setw(ndigits) << std::uppercase << value;
    return formatter.str();
}

/**
 * Reformats the specified decimal number string into a hex string
 * with the specified number of hex digits.  (The maximum number
 * of supported hex digits is 16, corresponding to a 64-bit number.)
 *
 * @throws an exception of the parsing of the decimalString into
 *         an integer fails.
 */
std::string format(const std::string& decimalString, const int& ndigits)
{
    if (ndigits > 8) {
        int64_t value = boost::lexical_cast<int64_t>(decimalString);
        return format(value, ndigits);
    }
    else {
        int32_t value = boost::lexical_cast<int32_t>(decimalString);
        return format(value, ndigits);
    }
}

/**
 * Parses the specified string into a 64-bit decimal value,
 * assuming that the string contains a numeric value in hex,
 * but does not do any checking that the hex string is valid.
 *
 * NOTE that if the input string is not valid hex, the resulting
 * integer value will essentially be a random number.
 */
int64_t unsafeParse(const std::string& hexString)
{
    std::istringstream iss(hexString);
    uint64_t u64Value;
    iss >> std::hex >> u64Value;
    return (int64_t) u64Value;
}

/**
 * Tests whether the specified string contains a valid number
 * in hex format.
 */
bool isValidHex(const std::string& hexString)
{
    // the logic used in this routine is to parse the input
    // value, regenerate the hex string, and verify that the
    // result matches what was input.  

    std::string inputString = boost::trim_copy(hexString);
    boost::to_upper(inputString);
    if (inputString.find("0X") != 0) {return false;}

    int numberOfHexDigits = inputString.length() - 2;
    int64_t value = unsafeParse(inputString);
    std::string resultString = format(value, numberOfHexDigits);
    boost::to_upper(resultString);

    return (resultString == inputString);
}

/**
 * Parses the specified hex string into a 32-bit decimal value.
 * Returns true if the parsing succeeded, and false if not.
 */
bool parse(const std::string& hexString, int32_t& decimalValue)
{
    if (! isValidHex(hexString)) {return false;}
    if (hexString.length() > 10) {return false;}

    std::istringstream iss(hexString);
    uint32_t u32Value;
    iss >> std::hex >> u32Value;
    decimalValue = (int32_t) u32Value;
    return true;
}

/**
 * Parses the specified hex string into a 64-bit decimal value.
 * Returns true if the parsing succeeded, and false if not.
 */
bool parse(const std::string& hexString, int64_t& decimalValue)
{
    if (! isValidHex(hexString)) {return false;}
    if (hexString.length() > 18) {return false;}

    std::istringstream iss(hexString);
    uint64_t u64Value;
    iss >> std::hex >> u64Value;
    decimalValue = (int64_t) u64Value;
    return true;
}

};  // end of namespace HexUtils
};  // end of namespace novadaq
