#include <NovaDAQUtilities/ProcessUtils.h>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>
#include <iostream>
#include <fstream>

namespace BFS = boost::filesystem;

namespace novadaq {

boost::mutex ProcessUtils::_childPidMutex;

/**
 * Fetches all descendant process IDs for the specified process.
 *
 * @param parentPid The parent process ID.
 * @param descendantPidList The resulting list of descendant process IDs.
 *
 * @return Status code (zero for success).
 */
int ProcessUtils::
getDescendantProcessIds(const int& parentPid,
                        std::vector<int>& descendantPidList) {
    std::vector<int> parentPidList;
    parentPidList.push_back(parentPid);
    return ProcessUtils::getDescendantProcessIds(parentPidList,
                                                 descendantPidList);
}

/**
 * Fetches all descendant process IDs for the specified list of processes.
 *
 * @param parentPidList The list of parent process IDs.
 * @param descendantPidList The resulting list of descendant process IDs.
 *
 * @return Status code (zero for success).
 */
int ProcessUtils::
getDescendantProcessIds(const std::vector<int>& parentPidList,
                        std::vector<int>& descendantPidList) {
    int status;
    status = ProcessUtils::getChildProcessIds(parentPidList,
                                              descendantPidList);
    if (status != 0) {return status;}

    if (descendantPidList.size() > 0) {
        std::vector<int> childPidList;
        status = ProcessUtils::
            getDescendantProcessIds(descendantPidList,
                                    childPidList);
        if (status != 0) {return status;}
        if (childPidList.size() > 0) {
            for (uint32_t idx = 0; idx < childPidList.size(); ++idx) {
                descendantPidList.push_back(childPidList[idx]);
            }
        }
    }

    return 0;
}

/**
 * Fetches the child process IDs for the specified process.
 *
 * @param parentPid The parent process ID.
 * @param childPidList The resulting list of child process IDs.
 *
 * @return Status code (zero for success).
 */
int ProcessUtils::getChildProcessIds(const int& parentPid,
                                     std::vector<int>& childPidList) {

    boost::mutex::scoped_lock lock(_childPidMutex);

    // there has to be a better general scheme than this...

    int thisPID = getpid();
    std::string tmpFile("/tmp/novadaq_processutils_" +
                        boost::lexical_cast<std::string>(thisPID) +
                        ".tmpout");

    std::string cmdString("ps h -o pid --ppid");
    cmdString .append(" " + boost::lexical_cast<std::string>(parentPid));
    cmdString .append(" 1>" + tmpFile);
    cmdString .append(" 2>/dev/null");

    int retcode = system(cmdString.c_str());
    if (retcode == 0) {
        std::ifstream tmpStream(tmpFile.c_str());
        if (tmpStream.fail()) {
            return -5;
        }
        else {
            int pid;
            tmpStream >> pid;
            while (! tmpStream.eof() && ! tmpStream.fail()) {
                childPidList.push_back(pid);
                tmpStream >> pid;
            }
            tmpStream.close();

            BFS::remove(tmpFile);
        }
    }

    return 0;
}

/**
 * Fetches the child process IDs for the specified list of processes.
 *
 * @param parentPidList The list of parent process IDs.
 * @param childPidList The resulting list of child process IDs.
 *
 * @return Status code (zero for success).
 */
int ProcessUtils::getChildProcessIds(const std::vector<int>& parentPidList,
                                     std::vector<int>& childPidList) {

    for (uint32_t idx = 0; idx < parentPidList.size(); ++idx) {
        int status =
            ProcessUtils::getChildProcessIds(parentPidList[idx],
                                             childPidList);
        if (status != 0) {return status;}
    }
    return 0;
}

};  // end of namespace novadaq
